# SPDX-FileCopyrightText: PhiBo DinoTools (2023)
# SPDX-License-Identifier: GPL-3.0-or-later

from pprint import pformat
from typing import List, Optional

import click
import nagiosplugin

from ..cli import cli
from ..context import BooleanContext, PerfdataScalarContext
from ..helper import logger, RouterOSVersion
from ..resource import RouterOSCheckResource


class SystemNtpClientResource(RouterOSCheckResource):
    name = "NTP"

    def __init__(
        self,
        cmd_options,
        check: nagiosplugin.Check,
        expected_servers: Optional[List[str]] = None,
        last_update_before_warning: Optional[float] = None,
        last_update_before_critical: Optional[float] = None,
        offset_warning: Optional[float] = None,
        offset_critical: Optional[float] = None,
        stratum_warning: Optional[int] = None,
        stratum_critical: Optional[int] = None,
    ):
        super().__init__(cmd_options=cmd_options, check=check)

        self._expected_servers = expected_servers
        self._offset_warning = offset_warning
        self._offset_critical = offset_critical
        self._last_update_before_warning = last_update_before_warning
        self._last_update_before_critical = last_update_before_critical
        self._stratum_warning = stratum_warning
        self._stratum_critical = stratum_critical

    def probe(self):
        logger.info("Fetching ntp client data ...")
        call = self.api.path(
            "/system/ntp/client"
        )

        results = tuple(call)

        result = results[0]
        logger.debug(f"Extracted values {pformat(result)}")

        self._routeros_metric_values += [
            {"name": "enabled", "type": bool},
        ]

        if not result["enabled"]:
            self._check.results.add(
                nagiosplugin.Result(
                    nagiosplugin.Critical,
                    "NTP Client not enabled"
                )
            )
            return self.get_routeros_metric_item(result)

        #: Current address of the server the devices gets its time from
        current_server_address: Optional[str] = None

        if self.routeros_version < RouterOSVersion("7"):
            metric_values = [
                {"name": "last-adjustment", "dst": "offset", "type": self.parse_routeros_time_duration, "uom": "s"},
                {"name": "last-update-before", "type": self.parse_routeros_time_duration, "uom": "s"},
            ]
            metric_value_names_not_found = []
            for metric_value in metric_values:
                if metric_value["name"] not in result:
                    metric_value_names_not_found.append(metric_value["name"])

            if len(metric_value_names_not_found) > 0:
                self._check.results.add(
                    nagiosplugin.Result(
                        nagiosplugin.state.Critical,
                        (
                            f"Looks like NTP client not running. "
                            f"Unable to find values for {', '.join(metric_value_names_not_found)}"
                        )
                    )
                )
            else:
                self._routeros_metric_values += metric_values
                self._check.add(
                    nagiosplugin.ScalarContext(
                        name="last-update-before",
                        warning=self._last_update_before_warning,
                        critical=self._last_update_before_critical,
                    ),
                    nagiosplugin.ScalarContext(
                        name="offset",
                        warning=f"-{self._offset_warning}:{self._offset_warning}" if self._offset_warning else None,
                        critical=f"-{self._offset_critical}:{self._offset_critical}" if self._offset_critical else None,
                    ),
                )
                if self._expected_servers:
                    current_server_address = result.get("last-update-from")
                    if current_server_address is None:
                        self._check.results.add(
                            nagiosplugin.Result(
                                nagiosplugin.state.Unknown,
                                "Unable to get address of server (last-update-from)"
                            )
                        )
        else:
            self._routeros_metric_values += [
                {"name": "freq-drift", "type": float},
                {"name": "synced-stratum", "dst": "stratum", "type": int},
                {"name": "system-offset", "dst": "offset", "type": lambda v: float(v) / 1000, "uom": "s"},
            ]
            self._check.add(
                PerfdataScalarContext(
                    name="freq-drift",
                ),
                nagiosplugin.ScalarContext(
                    name="offset",
                    warning=f"-{self._offset_warning}:{self._offset_warning}" if self._offset_warning else None,
                    critical=f"-{self._offset_critical}:{self._offset_critical}" if self._offset_critical else None,
                ),
                nagiosplugin.ScalarContext(
                    name="stratum",
                    warning=self._stratum_warning,
                    critical=self._stratum_critical,
                ),
            )
            if self._expected_servers:
                current_server_address = result.get("synced-server")
                if current_server_address is None:
                    self._check.results.add(
                        nagiosplugin.Result(
                            nagiosplugin.state.Unknown,
                            "Unable to get address of server (synced-server)"
                        )
                    )

        if current_server_address and current_server_address not in self._expected_servers:
            self._check.results.add(
                nagiosplugin.Result(
                    nagiosplugin.state.Warn,
                    (
                        f"Server '{current_server_address}' not in list of expected servers: "
                        f"{', '.join(self._expected_servers)}"
                    )
                )
            )

        return self.get_routeros_metric_item(result)


class SystemNtpClientSummary(nagiosplugin.Summary):
    def ok(self, results: List[nagiosplugin.Result]):
        messages: List[str] = []
        for result in results:
            if result.metric and result.metric.name == "stratum":
                messages.append(f"Stratum is {result.metric.value}")
            if result.metric and result.metric.name == "offset":
                messages.append(f"Offset is {result.metric.value:.2f}s")

        return ", ".join(messages)


@cli.command("system.ntp.client")
@click.option(
    "--last-update-before-warning",
    help=(
        "The time from the NTP server should at least be synchronised in the last N seconds. "
        "Default: 30 minutes = 1800 seconds "
        "Note: This is only available on RouterOS 6.x"
    ),
    type=float,
    default=60 * 30,
)
@click.option(
    "--last-update-before-critical",
    help=(
        "The time from the NTP server should at least be synchronised in the last N seconds. "
        "Default: 60 minutes = 3600 seconds "
        "Note: This is only available on RouterOS 6.x"
    ),
    type=float,
    default=60 * 60,
)
@click.option(
    "--offset-warning",
    help="Warning threshold for offset from the NTP server in seconds",
    type=float,
    default=10.0,
)
@click.option(
    "--offset-critical",
    help="Critical threshold for offset from the NTP server in seconds",
    type=float,
    default=30.0,
)
@click.option(
    "--stratum-warning",
    help=(
        "Check the stratum and report warning state if it does not match. "
        "Note: The stratum is only available on RouterOS 7.x"
    ),
    type=int,
)
@click.option(
    "--stratum-critical",
    help=(
        "Check the stratum and report critical state if it does not match. "
        "Note: The stratum is only available on RouterOS 7.x"
    ),
    type=int,
)
@click.option(
    "expected_servers",
    "--expected-server",
    multiple=True,
    help=(
        "Address of the ntp server we expect to get our time from. "
        "This must be the IPv4/IPv6 address and not the FQDN. "
        "It can be provided multiple times. "
        "Example: --expected-server 10.0.0.1 --expected-server 192.168.1.1"
    ),

)
@click.pass_context
@nagiosplugin.guarded
def system_clock(ctx, last_update_before_warning, last_update_before_critical, offset_warning, offset_critical,
                 stratum_warning, stratum_critical, expected_servers):
    """
    This command reads the information from /system/ntp/client to extract the required information.

    It checks if is the NTP client enabled, if the NTP server is reachable and if is the offset in the threshold.
    """
    check = nagiosplugin.Check()

    check.add(
        SystemNtpClientResource(
            cmd_options=ctx.obj,
            check=check,
            last_update_before_warning=last_update_before_warning,
            last_update_before_critical=last_update_before_critical,
            offset_warning=offset_warning,
            offset_critical=offset_critical,
            stratum_warning=stratum_warning,
            stratum_critical=stratum_critical,
            expected_servers=expected_servers,
        ),
        SystemNtpClientSummary(),
        BooleanContext(
            name="enabled",
        )
    )

    check.main(verbose=ctx.obj["verbose"])



# SPDX-FileCopyrightText: PhiBo DinoTools (2021)
# SPDX-License-Identifier: GPL-3.0-or-later
from datetime import date, datetime, time
from decimal import Decimal
from pprint import pformat
import re
import ssl
from typing import Any, Dict, List, Optional, Union

import librouteros
import librouteros.query
import nagiosplugin

from .helper import logger, RouterOSVersion
from .exeption import MissingValue


class RouterOSCheckResource(nagiosplugin.Resource):
    month_mapping: Dict[str, int] = {
        "jan": 1,
        "feb": 2,
        "mar": 3,
        "apr": 4,
        "may": 5,
        "jun": 6,
        "jul": 7,
        "aug": 8,
        "sep": 9,
        "oct": 10,
        "nov": 11,
        "dec": 12,
    }

    regex_datetime = re.compile(
        r"(?P<month>[a-z]{3})/(?P<day>\d+)/(?P<year>\d{4})\s+(?P<hour>\d+):(?P<minute>\d+):(?P<second>\d+)",
        flags=re.IGNORECASE
    )

    regex_date = re.compile(
        r"(?P<month>[a-z]{3})/(?P<day>\d+)/(?P<year>\d{4})",
        flags=re.IGNORECASE
    )

    regex_date_iso = re.compile(
        r"(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})",
        flags=re.IGNORECASE
    )

    regex_time = re.compile(
        r"(?P<hour>\d+):(?P<minute>\d+):(?P<second>\d+)",
        flags=re.IGNORECASE
    )

    def __init__(self, cmd_options: Dict[str, Any], check: nagiosplugin.Check):
        self._cmd_options = cmd_options
        self._check = check
        self._routeros_metric_values: List[Dict[str, Any]] = []
        self._routeros_version: Optional[RouterOSVersion] = None
        self._api: Optional[librouteros.api.Api] = None
        self.current_time = datetime.now()

    @property
    def api(self):
        if self._api is None:
            self._api = self.connect_api()

        return self._api

    @property
    def routeros_version(self):
        if self._routeros_version is None:
            if self._cmd_options["routeros_version"].strip().lower() == "auto":
                self._routeros_version = self._get_routeros_version()
            else:
                self._routeros_version = RouterOSVersion(self._cmd_options["routeros_version"].strip())

        return self._routeros_version

    @staticmethod
    def _calc_rate(
            cookie: nagiosplugin.Cookie,
            name: str,
            cur_value: int,
            elapsed_seconds: Optional[float],
            factor: int
    ) -> float:
        old_value: Optional[int] = cookie.get(name)
        cookie[name] = cur_value
        if old_value is None:
            raise MissingValue(f"Unable to find old value for '{name}'")
        if elapsed_seconds is None:
            raise MissingValue("Unable to get elapsed seconds")
        return (cur_value - old_value) / elapsed_seconds * factor

    def _connect_api(self) -> librouteros.api.Api:
        def wrap_socket(socket):
            server_hostname: Optional[str] = self._cmd_options["hostname"]
            if server_hostname is None:
                server_hostname = self._cmd_options["host"]
            return ssl_ctx.wrap_socket(socket, server_hostname=server_hostname)

        logger.info("Connecting to device ...")
        port = self._cmd_options["port"]
        extra_kwargs = {}
        if self._cmd_options["ssl"]:
            if port is None:
                port = 8729

            context_kwargs = {}
            if self._cmd_options["ssl_cafile"]:
                context_kwargs["cafile"] = self._cmd_options["ssl_cafile"]
            if self._cmd_options["ssl_capath"]:
                context_kwargs["capath"] = self._cmd_options["ssl_capath"]

            ssl_ctx = ssl.create_default_context(**context_kwargs)

            if self._cmd_options["ssl_force_no_certificate"]:
                ssl_ctx.check_hostname = False
                ssl_ctx.set_ciphers("ADH:@SECLEVEL=0")
            elif not self._cmd_options["ssl_verify"]:
                # We have do disable hostname check if we disable certificate verification
                ssl_ctx.check_hostname = False
                ssl_ctx.verify_mode = ssl.CERT_NONE
            elif not self._cmd_options["ssl_verify_hostname"]:
                ssl_ctx.check_hostname = False

            extra_kwargs["ssl_wrapper"] = wrap_socket
        else:
            if port is None:
                port = 8728

        api = librouteros.connect(
            host=self._cmd_options["host"],
            username=self._cmd_options["username"],
            password=self._cmd_options["password"],
            port=port,
            **extra_kwargs
        )
        return api

    @staticmethod
    def _convert_v6_list_to_v7(api_results) -> List[Dict[str, Any]]:
        result_items = []
        for name, value in api_results[0].items():
            result_items.append({
                "name": name,
                "value": value,
            })
        return result_items

    def _get_routeros_version(self) -> RouterOSVersion:
        call = self.api.path(
            "/system/resource"
        )
        results = tuple(call)
        result: Dict[str, str] = results[0]
        # version: 7.8 (stable)
        version_string = result["version"].partition(" ")[0]
        return RouterOSVersion(version_string)

    def connect_api(self) -> librouteros.api.Api:
        if self._api is None:
            self._api = self._connect_api()

        return self._api

    @classmethod
    def parse_routeros_date(cls, date_string: str) -> date:
        logger.debug(f"Parsing date string {date_string}")
        # Try iso date
        # Looks like they have switched date format in 7.11
        try:
            return date.fromisoformat(date_string)
        except ValueError as e:
            logger.debug(f"'{date_string}' is not in iso format: {str(e)}")

        # Try US date
        m = cls.regex_date.match(date_string)
        if m:
            return date(
                year=int(m.group("year")),
                month=cls.month_mapping[m.group("month").lower()],
                day=int(m.group("day"))
            )

        raise ValueError("Unable to parse datetime string")

    @classmethod
    def parse_routeros_date_time(cls, date_string: str, time_string: str) -> datetime:
        parsed_date = cls.parse_routeros_date(date_string=date_string)
        parsed_time = cls.parse_routeros_time(time_string=time_string)

        return datetime.combine(parsed_date, parsed_time)

    @classmethod
    def parse_routeros_datetime(cls, datetime_string: str) -> datetime:
        logger.debug(f"Parsing datetime string {datetime_string}")

        try:
            return datetime.fromisoformat(datetime_string)
        except ValueError as e:
            logger.debug(f"'{datetime_string}' is not in iso format: {str(e)}")

        m = cls.regex_datetime.match(datetime_string)
        if not m:
            raise ValueError("Unable to parse datetime string")

        return datetime(
            year=int(m.group("year")),
            month=cls.month_mapping[m.group("month").lower()],
            day=int(m.group("day")),
            hour=int(m.group("hour")),
            minute=int(m.group("minute")),
            second=int(m.group("second"))
        )

    @staticmethod
    def parse_routeros_speed(value_string: str) -> int:
        factors = {
            "": 1,
            "K": 1000,
            "M": 1000 * 1000,
            "G": 1000 * 1000 * 1000,
        }

        m = re.compile(r"(?P<value>\d+)(?P<factor>[A-Z]*)bps").match(value_string)
        if not m:
            raise ValueError(f"Unable to parse speed string: '{value_string}'")

        factor = factors.get(m.group("factor"))

        if factor is None:
            raise ValueError(f"Unable to parse element '{m.group()}' of speed string: '{value_string}'")

        return int(m.group("value")) * factor

    @classmethod
    def parse_routeros_time(cls, time_string: str) -> time:
        m = cls.regex_time.match(time_string)
        if not m:
            raise ValueError("Unable to parse datetime string")

        return time(
            hour=int(m.group("hour")),
            minute=int(m.group("minute")),
            second=int(m.group("second"))
        )

    @staticmethod
    def parse_routeros_time_duration(time_string: str) -> float:
        factors: Dict[str, Union[int, Decimal]] = {
            "us": Decimal(1e-6),
            "ms": Decimal(0.001),
            "s": 1,
            "m": 60,
            "h": 60 * 60,
            "d": 24 * 60 * 60,
            "w": 7 * 24 * 60 * 60,
        }

        value_is_negativ = time_string.startswith("-")

        seconds = Decimal(0)
        for m in re.compile(r"(?P<value>\d+)(?P<type>[a-z]+)").finditer(time_string):
            factor = factors.get(m.group("type"))
            if factor is None:
                raise ValueError(f"Unable to parse element '{m.group()}' of time string: '{time_string}'")
            seconds += int(m.group("value")) * factor

        seconds_float = float(round(seconds, 6))

        if value_is_negativ:
            return -seconds_float
        return seconds_float

    @staticmethod
    def prepare_override_values(override_values: List[str]) -> Dict[str, str]:
        results: Dict[str, str] = {}
        for override_value in override_values:
            name, _, value = override_value.partition(":")
            if value is None or value == "":
                logger.warning(f"Unable to parse override value for {name}")
            results[name] = value
        return results

    @staticmethod
    def prepare_thresholds(thresholds: List[str]) -> Dict[str, str]:
        results: Dict[str, str] = {}
        for threshold in thresholds:
            name, _, value = threshold.partition(":")
            if value is None or value == "":
                logger.warning(f"Unable to parse threshold for {name}")
            results[name] = value
        return results

    @staticmethod
    def prepare_regex_thresholds(thresholds: List[str]) -> Dict[re.Pattern, str]:
        results: Dict[re.Pattern, str] = {}
        for threshold in thresholds:
            name, _, value = threshold.partition(":")
            if value is None or value == "":
                logger.warning(f"Unable to parse threshold for {name}")
            results[re.compile(name)] = value
        return results

    def get_routeros_select_keys(self) -> List[librouteros.query.Key]:
        keys = []
        for metric_value in self._routeros_metric_values:
            keys.append(librouteros.query.Key(metric_value["name"]))
        return keys

    def get_routeros_metric_item(
        self, api_result: Dict[str, Any], name_prefix="", cookie=None
    ) -> List[nagiosplugin.Metric]:
        metrics = []

        elapsed_seconds = None
        if cookie:
            last_time_tuple = cookie.get("last_time")
            if isinstance(last_time_tuple, (list, tuple)):
                last_time = datetime(*last_time_tuple[0:6])
                delta_time = self.current_time - last_time
                elapsed_seconds = delta_time.total_seconds()

        #
        for metric_value in self._routeros_metric_values:
            metric_value_name = metric_value["name"]
            if metric_value_name not in api_result:
                if not metric_value.get("missing_ok", False):
                    self._check.results.add(
                        nagiosplugin.Result(
                            nagiosplugin.state.Unknown,
                            hint=f"Required metric with name '{metric_value_name}' not found in response"
                        )
                    )
                continue

            value = api_result[metric_value_name]
            metric_value_type = metric_value.get("type")
            if callable(metric_value_type):
                try:
                    value = metric_value_type(value)
                except ValueError as e:
                    logger.warning(f"Error parsing value with name {metric_value_name}", exc_info=True)
                    raise e

            value_factor = metric_value.get("factor")
            if value_factor is not None:
                if isinstance(value, bool):
                    logger.warning(f"Unable to use factor with bool metric {metric_value_name}", exc_info=True)
                    raise ValueError(f"Unable to use factor with bool metric {metric_value_name}")
                value = value * value_factor

            extra_kwargs = {}
            for n in ("min", "max", "uom"):
                if n in metric_value:
                    extra_kwargs[n] = metric_value[n]

            dst_value_name = metric_value.get("dst_value_name")
            if isinstance(dst_value_name, str):
                api_result[dst_value_name] = value

            if not metric_value.get("no_metric"):
                metrics.append(
                    nagiosplugin.Metric(
                        name=name_prefix + metric_value.get("dst", metric_value_name),
                        value=value,
                        **extra_kwargs,
                    )
                )

            if metric_value.get("rate"):
                try:
                    rate_value = self._calc_rate(
                        cookie=cookie,
                        name=metric_value_name,
                        cur_value=value,
                        elapsed_seconds=elapsed_seconds,
                        factor=metric_value.get("rate_factor", 1)
                    )
                    metrics.append(
                        nagiosplugin.Metric(
                            name=f"{name_prefix}{metric_value.get('dst', metric_value_name)}_rate",
                            value=rate_value,
                            uom=metric_value.get("rate_uom"),
                            min=metric_value.get("rate_min"),
                            max=metric_value.get("rate_max"),
                        )
                    )
                except MissingValue as e:
                    logger.debug(f"{e}", exc_info=e)

        if cookie:
            cookie["last_time"] = self.current_time.timetuple()

        return metrics

    def get_routeros_metrics(
        self, api_results: Union[List[Dict[str, Any]], Dict[str, Any]], name_prefix="", cookie=None
    ) -> List[nagiosplugin.Metric]:
        def get_api_result_by_name(api_results, name):
            for item in api_results:
                if name == item["name"]:
                    return item
            return None

        def new_api_result_item(api_results, item, ignore_if_exist=True):
            tmp_item = get_api_result_by_name(api_results, item["name"])
            if tmp_item is not None:
                api_results.append(item)
                return api_results

            if ignore_if_exist:
                return api_results

            raise ValueError("Duplicated entry")

        metrics = []

        elapsed_seconds = None
        if cookie:
            last_time_tuple = cookie.get("last_time")
            if isinstance(last_time_tuple, (list, tuple)):
                last_time = datetime(*last_time_tuple[0:6])
                delta_time = self.current_time - last_time
                elapsed_seconds = delta_time.total_seconds()

        if isinstance(api_results, dict):
            logger.debug(f"Extracted values {pformat(api_results)}")
            api_results = self._convert_v6_list_to_v7(api_results=api_results)

        #
        for metric_value in self._routeros_metric_values:
            metric_value_name = metric_value["name"]
            api_result = get_api_result_by_name(api_results, metric_value_name)

            if api_result is None:
                if not metric_value.get("missing_ok", False):
                    self._check.results.add(
                        nagiosplugin.Result(
                            nagiosplugin.state.Unknown,
                            hint=f"Required metric with name '{metric_value_name}' not found in response"
                        )
                    )
                continue

            value = api_result["value"]
            metric_value_type = metric_value.get("type")
            if callable(metric_value_type):
                try:
                    value = metric_value_type(value)
                except ValueError as e:
                    logger.warning(f"Error parsing value with name {metric_value_name}", exc_info=True)
                    raise e

            value = value * metric_value.get("factor", 1)

            extra_kwargs = {}
            for n in ("min", "max", "uom"):
                if n in metric_value:
                    extra_kwargs[n] = metric_value[n]

            dst_value_name = metric_value.get("dst_value_name")
            if isinstance(dst_value_name, str):
                api_results = new_api_result_item(
                    api_results,
                    {
                        "name": dst_value_name,
                        "value": value,
                    },
                    ignore_if_exist=True
                )

            if not metric_value.get("no_metric"):
                metrics.append(
                    nagiosplugin.Metric(
                        name=name_prefix + metric_value.get("dst", metric_value_name),
                        value=value,
                        **extra_kwargs,
                    )
                )

            if metric_value.get("rate"):
                try:
                    rate_value = self._calc_rate(
                        cookie=cookie,
                        name=metric_value_name,
                        cur_value=value,
                        elapsed_seconds=elapsed_seconds,
                        factor=metric_value.get("rate_factor", 1)
                    )
                    metrics.append(
                        nagiosplugin.Metric(
                            name=f"{name_prefix}{metric_value.get('dst', metric_value_name)}_rate",
                            value=rate_value,
                            uom=metric_value.get("rate_uom"),
                            min=metric_value.get("rate_min"),
                            max=metric_value.get("rate_max"),
                        )
                    )
                except MissingValue as e:
                    logger.debug(f"{e}", exc_info=e)

        if cookie:
            cookie["last_time"] = self.current_time.timetuple()

        return metrics



# SPDX-FileCopyrightText: PhiBo DinoTools (2021)
# SPDX-License-Identifier: GPL-3.0-or-later

from pprint import pformat
import re
from typing import Dict, List

import click
import librouteros
import librouteros.query
import nagiosplugin

from ..cli import cli
from ..helper import logger
from ..resource import RouterOSCheckResource


class SystemCpuResource(RouterOSCheckResource):
    name = "CPU"

    def __init__(
        self,
        cmd_options,
        check: nagiosplugin.Check,
        warning_values: List[str],
        critical_values: List[str],
        use_regex: bool
    ):
        super().__init__(cmd_options=cmd_options, check=check)

        self.values: Dict[str, float] = {}
        self.use_regex: bool = use_regex

        self.warning_values: Dict[str, str] = {}
        self.critical_values: Dict[str, str] = {}
        self.warning_regex_values: Dict[re.Pattern, str] = {}
        self.critical_regex_values: Dict[re.Pattern, str] = {}

        if self.use_regex:
            self.warning_regex_values = self.prepare_regex_thresholds(warning_values)
            self.critical_regex_values = self.prepare_regex_thresholds(critical_values)
        else:
            self.warning_values = self.prepare_thresholds(warning_values)
            self.critical_values = self.prepare_thresholds(critical_values)

    def probe(self):
        key_cpu_load = librouteros.query.Key("cpu-load")
        api = self._connect_api()

        logger.info("Fetching global data ...")
        call = api.path(
            "/system/resource"
        ).select(
            key_cpu_load
        )
        results = tuple(call)
        result = results[0]
        logger.debug(f"Extracted values {pformat(result)}")

        yield nagiosplugin.Metric(
            name="cpu-load",
            value=result["cpu-load"],
            uom="%",
            min=0,
            max=100,
        )

        logger.info("Fetching cpu data ...")
        call = api.path(
            "/system/resource/cpu"
        )
        results = tuple(call)
        logger.debug(f"Extracted values {pformat(results)}")

        for cpu in results:
            name = cpu["cpu"]
            for value_name_suffix in ("load", "irq", "disk"):
                value_name = f"{name}-{value_name_suffix}"
                if self.use_regex:
                    for regex, threshold in self.warning_regex_values.items():
                        if regex.match(value_name):
                            self.warning_values[value_name] = threshold
                            break

                    for regex, threshold in self.critical_regex_values.items():
                        if regex.match(value_name):
                            self.critical_values[value_name] = threshold
                            break

                self.values[value_name] = float(cpu[value_name_suffix])

        for name, value in self.values.items():
            self._check.add(nagiosplugin.ScalarContext(
                name=name,
                warning=self.warning_values.get(name),
                critical=self.critical_values.get(name),
            ))
            yield nagiosplugin.Metric(
                name=name,
                value=value,
                uom="%",
                min=0,
                max=100,
            )


class SystemCpuSummary(nagiosplugin.Summary):
    def ok(self, results: List[nagiosplugin.Result]):
        for result in results:
            if result.metric and result.metric.name == "cpu-load":
                return f"System load is {result.metric.value}%"

        return ""


@cli.command("system.cpu")
@click.option(
    "--load-warning",
    help="Warning threshold for global cpu load",
)
@click.option(
    "--load-critical",
    help="Critical threshold for global cpu load",
)
@click.option(
    "warning_values",
    "--value-warning",
    multiple=True,
    help=(
            "Set a warning threshold for a value. "
            "Example: If cpu1-load should be in the range of 10% to 20% you can set "
            "--value-warning cpu-load:10:200 "
            "Can be specified multiple times"
    )
)
@click.option(
    "critical_values",
    "--value-critical",
    multiple=True,
    help=(
        "Set a critical threshold for a value. "
        "Example: If cpu1-load should be in the range of 10% to 20% you can set "
        "--value-critical cpu-load:10:200 "
        "Can be specified multiple times"
    )
)
@click.option(
    "--regex",
    "use_regex",
    default=False,
    is_flag=True,
    help=(
        "Treat values from --value-warning and --value-critical as regex to find all matching values."
        "Example: Warn if cpu load of at least one CPU is above 80%: --value-warning 'cpu\\d+-load:80'"
    )
)
@click.pass_context
@nagiosplugin.guarded
def system_cpu(ctx, load_warning, load_critical, warning_values, critical_values, use_regex):
    """This command reads the information from /system/resource and /system/resource/cpu to extract the required
    information.
    """
    check = nagiosplugin.Check()

    check.add(
        SystemCpuResource(
            cmd_options=ctx.obj,
            check=check,
            warning_values=warning_values,
            critical_values=critical_values,
            use_regex=use_regex,
        ),
        nagiosplugin.ScalarContext(
            name="cpu-load",
            warning=load_warning,
            critical=load_critical,
        ),
        SystemCpuSummary(),
    )

    check.main(verbose=ctx.obj["verbose"])



# SPDX-FileCopyrightText: PhiBo DinoTools (2021)
# SPDX-License-Identifier: GPL-3.0-or-later

from routeros_check import helper


class TestRouterOSVersion:
    _CMP_EQ_ITEMS = (
        ("7.8", "7.8"),
        ("6.44beta41", "6.44beta41"),
        ("6.47.9", "6.47.9"),
        ("7.8", "7.8.0"),
    )
    _CMP_GT_ITEMS = (
        ("7.8", "6.47.9"),
    )
    _CMP_LT_ITEMS = (
        ("6.47.9", "7.8"),
    )
    _CMP_NE_ITEMS = (
        ("7.8", "6.44beta41"),
        ("6.44beta41", "6.47.9"),
        ("6.47.9", "7.8"),
        ("7.8", "7.8.1"),
        ("7.8", "7.9"),
        ("7.8.1", "7.7.1"),
        ("6.8.1", "7.8.1"),
    )

    def test_compare_routeros_version_eq(self):
        cmp_items = self._CMP_EQ_ITEMS
        for item in cmp_items:
            assert helper.RouterOSVersion(item[0]) == helper.RouterOSVersion(item[1])

        cmp_items = self._CMP_NE_ITEMS + self._CMP_GT_ITEMS + self._CMP_LT_ITEMS
        for item in cmp_items:
            assert not helper.RouterOSVersion(item[0]) == helper.RouterOSVersion(item[1])

    def test_compare_routeros_version_ge(self):
        cmp_items = self._CMP_GT_ITEMS + self._CMP_EQ_ITEMS
        for item in cmp_items:
            assert helper.RouterOSVersion(item[0]) >= helper.RouterOSVersion(item[1])

        cmp_items = self._CMP_LT_ITEMS
        for item in cmp_items:
            assert not helper.RouterOSVersion(item[0]) >= helper.RouterOSVersion(item[1])

    def test_compare_routeros_version_gt(self):
        cmp_items = self._CMP_GT_ITEMS
        for item in cmp_items:
            assert helper.RouterOSVersion(item[0]) > helper.RouterOSVersion(item[1])

        cmp_items = self._CMP_LT_ITEMS + self._CMP_EQ_ITEMS
        for item in cmp_items:
            assert not helper.RouterOSVersion(item[0]) > helper.RouterOSVersion(item[1])

    def test_compare_routeros_version_le(self):
        cmp_items = self._CMP_LT_ITEMS + self._CMP_EQ_ITEMS
        for item in cmp_items:
            assert helper.RouterOSVersion(item[0]) <= helper.RouterOSVersion(item[1])

        cmp_items = self._CMP_GT_ITEMS
        for item in cmp_items:
            assert not helper.RouterOSVersion(item[0]) <= helper.RouterOSVersion(item[1])

    def test_compare_routeros_version_lt(self):
        cmp_items = self._CMP_LT_ITEMS
        for item in cmp_items:
            assert helper.RouterOSVersion(item[0]) < helper.RouterOSVersion(item[1])

        cmp_items = self._CMP_EQ_ITEMS + self._CMP_GT_ITEMS
        for item in cmp_items:
            assert not helper.RouterOSVersion(item[0]) < helper.RouterOSVersion(item[1])

    def test_humanize_time_with_int(self):
        assert helper.humanize_time(10) == "10s"
        assert helper.humanize_time(75) == "1m 15s"
        assert helper.humanize_time(3601) == "1h 0m 1s"
        assert helper.humanize_time(
            60 * 60 * 24 +
            60 * 60 * 1 +
            60 * 13 +
            44
        ) == "1d 1h 13m 44s"

    def test_humanize_time_with_timedelta(self):
        from datetime import timedelta
        assert helper.humanize_time(
            timedelta(
                days=2,
                hours=12,
                minutes=33,
                seconds=56
            )
        ) == "2d 12h 33m 56s"

    def test_parse_routeros_version(self):
        a = helper.RouterOSVersion("7.8")
        assert a.major == 7
        assert a.minor == 8
        assert a.patch == 0

        b = helper.RouterOSVersion("6.44beta41")
        assert b.major == 6
        assert b.minor == 44
        assert b.patch == 0

        c = helper.RouterOSVersion("6.47.9")
        assert c.major == 6
        assert c.minor == 47
        assert c.patch == 9



# SPDX-FileCopyrightText: PhiBo DinoTools (2021)
# SPDX-License-Identifier: GPL-3.0-or-later

import re
from typing import Dict, List, Set

import click
import nagiosplugin

from ..cli import cli
from ..context import BooleanContext
from ..helper import logger, RouterOSVersion
from ..resource import RouterOSCheckResource


class SystemPsuResource(RouterOSCheckResource):
    name = "PSU"

    def __init__(
        self, cmd_options, check: nagiosplugin.Check, warning_values: List[str], critical_values: List[str],
        no_psu_ok: bool,
    ):
        super().__init__(cmd_options=cmd_options, check=check)

        self.psu_names: Set[str] = set()
        self.psu_states: Dict[str, str] = {}
        self.psu_values: Dict[str, float] = {}
        self.warning_values = self._prepare_thresholds(warning_values)
        self.critical_values = self._prepare_thresholds(critical_values)
        self.no_psu_ok = no_psu_ok

        self._fetch_data()

    def _fetch_data(self):
        logger.info("Fetching data ...")
        call = self.api.path(
            "/system/health"
        )
        api_results = tuple(call)
        if self.routeros_version < RouterOSVersion("7"):
            api_result_items = []
            for name, value in api_results[0].items():
                api_result_items.append({
                    "name": name,
                    "value": value,
                })
        else:
            api_result_items = api_results

        regex_name = re.compile(r"(?P<name>psu\d+)-(?P<type>(state|current|voltage))")
        for api_result_item in api_result_items:
            m = regex_name.match(api_result_item["name"])
            if not m:
                continue

            self.psu_names.add(m.group("name"))

            if m.group("type") in ("current", "voltage"):
                self.psu_values[api_result_item["name"]] = float(api_result_item["value"])

            if m.group("type") == "state":
                self.psu_states[m.group("name")] = api_result_item["value"]

        if not self.no_psu_ok and len(self.psu_values) == 0 and len(self.psu_states) == 0:
            self._check.results.add(
                nagiosplugin.Result(
                    nagiosplugin.state.Unknown,
                    hint="No PSU values and stats found"
                )
            )

    @staticmethod
    def _prepare_thresholds(thresholds: List[str]):
        results = {}
        for threshold in thresholds:
            name, _, value = threshold.partition(":")
            if value is None or value == "":
                logger.warning(f"Unable to parse threshold for {name}")
            results[name] = value
        return results

    def probe(self):
        for name, value in self.psu_values.items():
            self._check.add(nagiosplugin.ScalarContext(
                name=name,
                warning=self.warning_values.get(name),
                critical=self.critical_values.get(name),
            ))
            yield nagiosplugin.Metric(
                name=name,
                value=value,
            )
        for name, value in self.psu_states.items():
            value_name = f"{name}-state-ok"
            self._check.add(
                BooleanContext(value_name)
            )
            if value != "ok":
                self._check.results.add(
                    nagiosplugin.Result(
                        nagiosplugin.state.Warn,
                        hint=f"PSU: {name} state {value}"
                    )
                )
            yield nagiosplugin.Metric(
                name=value_name,
                value=(value == "ok")
            )


@cli.command("system.psu")
@click.option(
    "warning_values",
    "--value-warning",
    multiple=True,
    help=(
        "Set a warning threshold for a value. "
        "Example: If psu1-voltage should be in the range of 12-12.1V you can set --value-warning psu1-voltage:12:12.1 "
        "Can be specified multiple times"
    )
)
@click.option(
    "critical_values",
    "--value-critical",
    multiple=True,
    help=(
        "Set a critical threshold for a value. "
        "Example: If psu1-voltage should be in the range of 12-12.1V you can set --value-critical psu1-voltage:12:12.1 "
        "Can be specified multiple times"
    )
)
@click.option(
    "--no-psu-ok",
    is_flag=True,
    default=False,
    help="The check will be unknown if not at least one psu stat or value is available. Set this to ignore this."
)
@click.option(
    "expected_psu_names",
    "--expect-psu",
    multiple=True,
    default=[],
    help="Name of the PSU to expect at least one value or state. Can be specified multiple times."
)
@click.pass_context
@nagiosplugin.guarded
def system_psu(ctx, warning_values, critical_values, no_psu_ok, expected_psu_names):
    """Check the power supply units (PSU)"""
    check = nagiosplugin.Check()

    psu_resource = SystemPsuResource(
        cmd_options=ctx.obj,
        check=check,
        warning_values=warning_values,
        critical_values=critical_values,
        no_psu_ok=no_psu_ok,
    )
    check.add(psu_resource)

    check.results.add(
        nagiosplugin.Result(
            nagiosplugin.state.Ok,
            hint=f"Looks like all PSU work like expected: {', '.join(psu_resource.psu_names)}"
        )
    )
    if len(expected_psu_names) > 0:
        missing_psu_names = []
        for psu_name in expected_psu_names:
            if psu_name not in psu_resource.psu_names:
                missing_psu_names.append(psu_name)

        if len(missing_psu_names) > 0:
            check.results.add(
                nagiosplugin.Result(
                    nagiosplugin.state.Warn,
                    hint=f"Expected PSU(s) not found: {', '.join(missing_psu_names)}"
                )
            )

    check.main(verbose=ctx.obj["verbose"])



# SPDX-FileCopyrightText: PhiBo DinoTools (2021)
# SPDX-License-Identifier: GPL-3.0-or-later

import re
from typing import Dict, List, Set

import click
import nagiosplugin

from ..cli import cli
from ..helper import logger, RouterOSVersion
from ..resource import RouterOSCheckResource


class SystemFanResource(RouterOSCheckResource):
    name = "FAN"

    def __init__(
        self,
        cmd_options,
        check: nagiosplugin.Check,
        warning_values: List[str],
        critical_values: List[str],
        use_regex: bool
    ):
        super().__init__(cmd_options=cmd_options, check=check)

        self.fan_names: Set[str] = set()
        self.fan_values: Dict[str, int] = {}
        self.use_regex: bool = use_regex

        self.warning_values: Dict[str, str] = {}
        self.critical_values: Dict[str, str] = {}
        self.warning_regex_values: Dict[re.Pattern, str] = {}
        self.critical_regex_values: Dict[re.Pattern, str] = {}

        if self.use_regex:
            self.warning_regex_values = self.prepare_regex_thresholds(warning_values)
            self.critical_regex_values = self.prepare_regex_thresholds(critical_values)
        else:
            self.warning_values = self.prepare_thresholds(warning_values)
            self.critical_values = self.prepare_thresholds(critical_values)

        self._fetch_data()

    def _fetch_data(self):
        logger.info("Fetching data ...")
        call = self.api.path(
            "/system/health"
        )
        api_results = tuple(call)
        if self.routeros_version < RouterOSVersion("7"):
            api_result_items = []
            for name, value in api_results[0].items():
                api_result_items.append({
                    "name": name,
                    "value": value,
                })
        else:
            api_result_items = api_results

        regex_name = re.compile(r"(?P<name>fan\d+)-(?P<type>(speed))")
        for item in api_result_items:
            m = regex_name.match(item["name"])
            if not m:
                continue

            if self.use_regex:
                for regex, threshold in self.warning_regex_values.items():
                    if regex.match(item["name"]):
                        self.warning_values[item["name"]] = threshold
                        break

                for regex, threshold in self.critical_regex_values.items():
                    if regex.match(item["name"]):
                        self.critical_values[item["name"]] = threshold
                        break

            if m.group("type") in ("speed",):
                self.fan_values[item["name"]] = int(item["value"])

            self.fan_names.add(m.group("name"))

    def probe(self):
        for name, value in self.fan_values.items():
            self._check.add(nagiosplugin.ScalarContext(
                name=name,
                warning=self.warning_values.get(name),
                critical=self.critical_values.get(name),
            ))
            yield nagiosplugin.Metric(
                name=name,
                value=value,
            )


@cli.command("system.fan")
@click.option(
    "warning_values",
    "--value-warning",
    multiple=True,
    help=(
        "Set a warning threshold for a value. "
        "Example: If fan1-speed should be in the range of 4000 to 5000 you can set "
        "--value-warning fan1-speed:4000:5000 "
        "Can be specified multiple times"
    )
)
@click.option(
    "critical_values",
    "--value-critical",
    multiple=True,
    help=(
        "Set a critical threshold for a value. "
        "Example: If fan1-speed should be in the range of 4000 to 5000 you can set "
        "--value-critical fan1-speed:4000:5000 "
        "Can be specified multiple times"
    )
)
@click.option(
    "--regex",
    "use_regex",
    default=False,
    is_flag=True,
    help="Treat values from --value-warning and --value-critical as regex to find all matching values"
)
@click.option(
    "--no-fan-ok",
    is_flag=True,
    default=False,
    help="The check will be unknown if no fan is available. Provide this option to ignore this."
)
@click.option(
    "expected_names",
    "--expect-fan",
    multiple=True,
    default=[],
    help="Name of the fan to expect. Can be specified multiple times."
)
@click.pass_context
@nagiosplugin.guarded
def system_fan(ctx, warning_values, critical_values, use_regex, no_fan_ok, expected_names):
    check = nagiosplugin.Check()

    fan_resource = SystemFanResource(
        cmd_options=ctx.obj,
        check=check,
        warning_values=warning_values,
        critical_values=critical_values,
        use_regex=use_regex,
    )
    check.add(fan_resource)

    check.results.add(
        nagiosplugin.Result(
            nagiosplugin.state.Ok,
            hint=f"Looks like all fans work as expected: {', '.join(sorted(fan_resource.fan_names))}"
        )
    )
    if len(fan_resource.fan_names) == 0 and not no_fan_ok:
        check.results.add(
            nagiosplugin.Result(
                nagiosplugin.state.Unknown,
                hint="No FANs found"
            )
        )

    if len(expected_names) > 0:
        missing_names = []
        for name in expected_names:
            if name not in fan_resource.fan_names:
                missing_names.append(name)

        if len(missing_names) > 0:
            check.results.add(
                nagiosplugin.Result(
                    nagiosplugin.state.Warn,
                    hint=f"Expected FAN(s) not found: {', '.join(missing_names)}"
                )
            )

    check.main(verbose=ctx.obj["verbose"])



# SPDX-FileCopyrightText: PhiBo DinoTools (2021)
# SPDX-License-Identifier: GPL-3.0-or-later

import click
import nagiosplugin

from ..cli import cli
from ..helper import logger, RouterOSVersion
from ..resource import RouterOSCheckResource


class SystemPowerResource(RouterOSCheckResource):
    name = "Power"

    def __init__(
        self,
        cmd_options,
        check: nagiosplugin.Check,
    ):
        super().__init__(cmd_options=cmd_options, check=check)

        self._routeros_metric_values = [
            {"name": "power-consumption", "type": float},
        ]

    def probe(self):
        logger.info("Fetching data ...")
        call = self.api.path(
            "/system/health"
        )
        if self.routeros_version < RouterOSVersion("7"):
            call = call.select(
                *self.get_routeros_select_keys()
            )
            api_result_items = tuple(call)
            api_result_items = self._convert_v6_list_to_v7(api_result_items)
        else:
            api_result_items = tuple(call)

        result_metrics = self.get_routeros_metrics(api_result_items)
        if len(result_metrics) == 0:
            self._check.results.add(
                nagiosplugin.Result(
                    nagiosplugin.state.Unknown,
                    hint="Power consumption not found."
                )
            )
        return result_metrics


@cli.command("system.power")
@click.option(
    "--warning",
    help="Warning threshold for total power consumption",
)
@click.option(
    "--critical",
    help="Critical threshold for total power consumption",
)
@click.pass_context
@nagiosplugin.guarded
def system_power(ctx, warning, critical):
    """Check the total power consumption of a device. This might not be available on all devices"""
    check = nagiosplugin.Check()

    check.add(
        SystemPowerResource(
            cmd_options=ctx.obj,
            check=check,
        ),
        nagiosplugin.ScalarContext(
            "power-consumption",
            warning=warning,
            critical=critical,
            fmt_metric="Power consumption {value}W",
        ),
    )

    check.main(verbose=ctx.obj["verbose"])



# SPDX-FileCopyrightText: PhiBo DinoTools (2023)
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import datetime
from pprint import pformat
from typing import List

import click
import nagiosplugin

from ..cli import cli
from ..context import SimplePositiveFloatContext
from ..helper import logger
from ..resource import RouterOSCheckResource


class SystemClockResource(RouterOSCheckResource):
    name = "CLOCK"

    def __init__(
        self,
        cmd_options,
        check: nagiosplugin.Check,
    ):
        super().__init__(cmd_options=cmd_options, check=check)

    def probe(self):
        api = self._connect_api()

        logger.info("Fetching clock data ...")
        call = api.path(
            "/system/clock"
        )

        results = tuple(call)

        result = results[0]
        logger.debug(f"Extracted values {pformat(result)}")

        device_datetime = self.parse_routeros_date_time(result["date"], result["time"])

        device_timediff = datetime.now() - device_datetime

        yield nagiosplugin.Metric(
            name="time-diff",
            value=device_timediff.total_seconds(),
            uom="s",
        )


class SystemClockSummary(nagiosplugin.Summary):
    def ok(self, results: List[nagiosplugin.Result]):
        for result in results:
            if result.metric and result.metric.name == "time-diff":
                return f"Time diff is {result.metric.value:.2f}s"

        return ""


@cli.command("system.clock")
@click.option(
    "--warning",
    help="Warning threshold for time diff in seconds",
    type=float,
)
@click.option(
    "--critical",
    help="Critical threshold for time diff in seconds",
    type=float,
)
@click.pass_context
@nagiosplugin.guarded
def system_clock(ctx, warning, critical):
    """This command reads the information from /system/clock to extract the required information."""
    check = nagiosplugin.Check()

    check.add(
        SystemClockResource(
            cmd_options=ctx.obj,
            check=check,
        ),
        SimplePositiveFloatContext(
            name="time-diff",
            warning=warning,
            critical=critical,
            fmt_metric="Time diff is {valueunit}",
        ),
        SystemClockSummary(),
    )

    check.main(verbose=ctx.obj["verbose"])



# SPDX-FileCopyrightText: PhiBo DinoTools (2021)
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import timedelta
import importlib
import logging
import os
import re
from typing import List, Optional, Union

logger = logging.getLogger('nagiosplugin')


REGEX_VERSION_PATTERN = r"""
    (?P<release>[0-9]+(?:\.[0-9]+)*)
    (?P<pre>
        [-_\.]?
        (?P<pre_type>(a|b|c|rc|alpha|beta|pre|preview))
        [-_\.]?
        (?P<pre_serial>[0-9]+)?
    )?
"""


class RouterOSVersion:
    def __init__(self, version_string: str):
        regex = re.compile(r"^\s*" + REGEX_VERSION_PATTERN + r"\s*$", re.VERBOSE | re.IGNORECASE)
        m = regex.match(version_string)
        if not m:
            raise ValueError(f"Unable to parse version string: '{version_string}'")

        self.release = tuple([int(v) for v in m.group("release").split(".")])
        # At the moment we don't handle the pre releases like alpha, beta or rc
        # We should try to work with major and minor version

        self._cmp_attribute_names = ("major", "minor", "patch")

    def __eq__(self, other):
        for attr_name in self._cmp_attribute_names:
            if getattr(self, attr_name) != getattr(other, attr_name):
                return False

        return True

    def __ge__(self, other):
        return self > other or self == other

    def __gt__(self, other):
        for attr_name in self._cmp_attribute_names:
            if getattr(self, attr_name) > getattr(other, attr_name):
                return True
            if getattr(self, attr_name) < getattr(other, attr_name):
                return False

        return False

    def __le__(self, other):
        return self < other or self == other

    def __lt__(self, other):
        for attr_name in self._cmp_attribute_names:
            if getattr(self, attr_name) < getattr(other, attr_name):
                return True
            if getattr(self, attr_name) > getattr(other, attr_name):
                return False

        return False

    def __repr__(self):
        return f"{self.__class__.__name__}('{self}')"

    def __str__(self):
        return f"{'.'.join([str(v) for v in self.release])}"

    @property
    def major(self) -> int:

        return self.release[0] if len(self.release) >= 1 else 0

    @property
    def minor(self) -> int:
        return self.release[1] if len(self.release) >= 2 else 0

    @property
    def patch(self) -> int:
        return self.release[2] if len(self.release) >= 3 else 0


def escape_filename(value):
    value = re.sub(r"[^\w\s-]", "_", value).strip().lower()
    return re.sub(r"[-\s]+", '-', value)


def humanize_time(time: Union[float, int, timedelta]):
    interval_mappings = {
        1: {
            "short": "s",
        },
        60: {
            "short": "m",
        },
        3600: {
            "short": "h",
        },
        86400: {
            "short": "d",
        }
    }

    if isinstance(time, timedelta):
        time = time.total_seconds()

    intervals: List[int] = sorted(interval_mappings.keys())
    intervals.reverse()

    results: List[str] = []
    for interval in intervals:
        v = int(time // interval)
        if v or len(results):
            results.append(f"{v}{interval_mappings[interval]['short']}")
        time = time % interval

    return " ".join(results)


def load_modules(pkg_names: Optional[List] = None):
    if pkg_names is None:
        pkg_names = [".check"]
    for base_pkg_name in pkg_names:
        logger.debug("Base package name: %s", base_pkg_name)
        base_pkg = importlib.import_module(base_pkg_name, package=__package__)

        logger.debug("Base package: %s", base_pkg)

        path = base_pkg.__path__[0]
        logger.debug("Base path: %s", path)

        for filename in os.listdir(path):
            if filename == "__init__.py":
                continue

            pkg_name = None
            if os.path.isdir(os.path.join(path, filename)) and \
                    os.path.exists(os.path.join(path, filename, "__init__.py")):
                pkg_name = filename

            if filename[-3:] == '.py':
                pkg_name = filename[:-3]

            if pkg_name is None:
                continue

            mod_name = "{}.{}".format(base_pkg_name, pkg_name)
            try:
                importlib.import_module(mod_name, package=__package__)
                logger.info("Loaded '%s' successfully", mod_name)
            except ImportError:
                logger.warning("Unable to load: '%s'", mod_name)
                logger.debug("An error occurred while importing '%s'", mod_name, exc_info=True)



# SPDX-FileCopyrightText: PhiBo DinoTools (2021)
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any, Dict, List, Optional

import click
import librouteros
import librouteros.query
import nagiosplugin

from ..cli import cli
from ..context import BooleanContext
from ..helper import humanize_time, logger, RouterOSVersion
from ..resource import RouterOSCheckResource


class RoutingOSPFNeighborResource(RouterOSCheckResource):
    name = "OSPF NEIGHBOR"

    def __init__(
            self,
            cmd_options: Dict[str, Any],
            check: nagiosplugin.Check,
            instance: str,
            router_id: str,
            area: Optional[str] = None
    ):
        super().__init__(cmd_options=cmd_options, check=check)

        self.area = area
        self.instance = instance
        self.router_id = router_id

        self.state: Optional[str] = None

        self._routeros_metric_values = [
            {
                "name": "adjacency",
                "type": self.parse_routeros_time_duration,
                "min": 0,
                "uom": "s",
                # on some devices, only available if state=Full
                "missing_ok": True,
            },
            {
                "name": "state",
                "type": None,
            },
            {
                "name": "state-changes",
                "dst": "state_changes",
                "type": int
            },
        ]
        if self.routeros_version < RouterOSVersion("7"):
            self._routeros_metric_values += [
                {"name": "priority", "type": int},
                {"name": "ls-retransmits", "dst": "ls_retransmits", "type": int},
                {"name": "ls-requests", "dst": "ls_requests", "type": int},
                {"name": "db-summaries", "dst": "db_summaries", "type": int},
            ]

    def probe(self):
        # ToDo: Only available in v7.x
        key_area = librouteros.query.Key("area")
        key_instance = librouteros.query.Key("instance")
        key_router_id = librouteros.query.Key("router-id")

        logger.info("Fetching data ...")

        select_keys = [
            key_instance,
            key_router_id,
        ] + self.get_routeros_select_keys()

        if self.routeros_version >= RouterOSVersion("7"):
            select_keys.append(key_area)

        where = [
            key_instance == self.instance,
            key_router_id == self.router_id,
        ]

        if self.area is not None:
            if self.routeros_version >= RouterOSVersion("7"):
                where.append(key_area == self.area)
            else:
                logger.warning("The area selector is only available on RouterOS 7.x")

        call = self.api.path(
            "/routing/ospf/neighbor"
        ).select(
            *select_keys
        ).where(
            *where
        )
        results = tuple(call)
        if len(results) == 0:
            return nagiosplugin.Metric(
                name="state",
                value=None
            )

        result = results[0]

        return self.get_routeros_metric_item(result)


class RoutingOSPFNeighborState(BooleanContext):
    def evaluate(self, metric, resource: RoutingOSPFNeighborResource):
        if metric.value is None:
            if resource.area is None:
                hint = f"Neighbor for instance '{resource.instance}' and router-id '{resource.router_id}' not found"
            else:
                hint = (
                    f"Neighbor for area '{resource.area}', instance '{resource.instance}' and "
                    f"router-id '{resource.router_id}' not found"
                )
            return nagiosplugin.Result(
                state=nagiosplugin.state.Critical,
                hint=hint,
                metric=metric,
            )
        elif metric.value in ("Down",):
            return self.result_cls(
                state=nagiosplugin.state.Critical,
                hint="Link to neighbor down",
                metric=metric,
            )
        elif metric.value in ("Full",):
            return self.result_cls(
                state=nagiosplugin.state.Ok,
                hint="Communicating with neighbor",
                metric=metric,
            )
        else:
            return self.result_cls(
                state=nagiosplugin.state.Warn,
                hint=f"Link to neighbor not fully up, state: {metric.value}",
                metric=metric,
            )


class RoutingOSPFNeighborSummary(nagiosplugin.Summary):
    def ok(self, results: List[nagiosplugin.Result]):
        result_parts: List[str] = []
        for result in results:
            if not result.metric:
                continue

            if result.metric.name == "adjacency":
                result_parts.append(f"Adjacency: {humanize_time(result.metric.value)}")

            if result.metric.name == "state":
                result_parts.append(f"State: {result.metric.value}")

        return " ".join(result_parts)


@cli.command("routing.ospf.neighbors")
@click.option(
    "--area",
    help="The area the neighbor router belongs to (only supported on RouterOS v7.x",
)
@click.option(
    "--instance",
    required=True,
    help="The name of the OSPF instance",
)
@click.option(
    "--router-id",
    required=True,
    help="The ID of the neighbor router",
)
@click.pass_context
def routing_ospf_neighbors(ctx, area, instance, router_id):
    """Check the state of an OSPF neighbor"""
    check = nagiosplugin.Check()

    check.add(
        RoutingOSPFNeighborResource(
            cmd_options=ctx.obj,
            area=area,
            instance=instance,
            router_id=router_id,
        ),
        nagiosplugin.ScalarContext("priority"),
        nagiosplugin.ScalarContext("adjacency"),
        nagiosplugin.ScalarContext("state_changes"),
        nagiosplugin.ScalarContext("ls_retransmits"),
        nagiosplugin.ScalarContext("ls_requests"),
        nagiosplugin.ScalarContext("db_summaries"),
        RoutingOSPFNeighborState("state"),
        RoutingOSPFNeighborSummary(),
    )

    check.main(verbose=ctx.obj["verbose"])



# SPDX-FileCopyrightText: PhiBo DinoTools (2021)
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import List

import click
import librouteros
import librouteros.query
import nagiosplugin

from ..cli import cli
from ..context import ScalarPercentContext
from ..helper import logger
from ..resource import RouterOSCheckResource


class SystemMemoryResource(RouterOSCheckResource):
    name = "MEMORY"

    def __init__(self, cmd_options, check: nagiosplugin.Check):
        super().__init__(cmd_options=cmd_options, check=check)

        self.memory_total = None

    def probe(self):
        api = self._connect_api()

        logger.info("Fetching data ...")
        call = api.path(
            "/system/resource"
        ).select(
            librouteros.query.Key("free-memory"),
            librouteros.query.Key("total-memory")
        )
        results = tuple(call)
        result = results[0]

        memory_free = result["free-memory"]
        self.memory_total = result["total-memory"]

        yield nagiosplugin.Metric(
            name="free",
            value=memory_free,
            uom="B",
            min=0,
            max=self.memory_total,
        )

        yield nagiosplugin.Metric(
            name="used",
            value=self.memory_total - memory_free,
            uom="B",
            min=0,
            max=self.memory_total,
        )


class SystemMemorySummary(nagiosplugin.summary.Summary):
    def __init__(self, result_names: List[str]):
        super().__init__()
        self._result_names = result_names

    def ok(self, results):
        msgs = []
        for result_name in self._result_names:
            msgs.append(str(results[result_name]))
        return " ".join(msgs)


@cli.command("system.memory")
@click.option(
    "--used/--free",
    is_flag=True,
    default=True,
    help="Set if used or free memory should be checked. (Default: used)",
)
@click.option(
    "--warning",
    required=True,
    help="Warning threshold in % or MB. Example (20% oder 20 = 20MB)",
)
@click.option(
    "--critical",
    required=True,
    help="Critical threshold in % or MB. Example (20% oder 20 = 20MB)",
)
@click.pass_context
@nagiosplugin.guarded
def system_memory(ctx, used, warning, critical):
    check = nagiosplugin.Check()

    check.add(
        SystemMemoryResource(
            cmd_options=ctx.obj,
            check=check,
        ),
    )

    if used:
        check.add(nagiosplugin.ScalarContext(
            name="free",
        ))
        check.add(ScalarPercentContext(
            name="used",
            total_name="memory_total",
            warning=warning,
            critical=critical
        ))
    else:
        check.add(ScalarPercentContext(
            name="free",
            total_name="memory_total",
            warning=f"{warning}:",
            critical=f"{critical}:"
        ))
        check.add(nagiosplugin.ScalarContext(
            name="used",
        ))

    check.add(SystemMemorySummary(
        result_names=["used"] if used else ["free"]
    ))

    check.main(verbose=ctx.obj["verbose"])



# SPDX-FileCopyrightText: PhiBo DinoTools (2021)
# SPDX-License-Identifier: GPL-3.0-or-later

import click
import librouteros
import librouteros.query
import nagiosplugin

from ..cli import cli
from ..context import BooleanContext
from ..helper import logger
from ..resource import RouterOSCheckResource


class InterfaceVrrpCheck(RouterOSCheckResource):
    name = "VRRP"

    def __init__(self, cmd_options, check: nagiosplugin.Check, name, master_must):
        super().__init__(cmd_options=cmd_options, check=check)

        self._name = name
        self.backup = None
        self.disabled = None
        self.enabled = None
        self.invalid = None
        self.master = None
        self.master_must = master_must
        self.running = None

    def probe(self):
        key_name = librouteros.query.Key("name")
        api = self._connect_api()

        logger.info("Fetching data ...")
        call = api.path(
            "/interface/vrrp"
        ).select(
            key_name,
            librouteros.query.Key("backup"),
            librouteros.query.Key("disabled"),
            librouteros.query.Key("invalid"),
            librouteros.query.Key("master"),
            librouteros.query.Key("running"),
        ).where(
            key_name == self._name
        )
        results = tuple(call)
        result = results[0]

        self.disabled = result["disabled"]
        self.enabled = not self.disabled

        yield nagiosplugin.Metric(
            name="disabled",
            value=self.disabled,
        )

        if self.enabled:
            for n in ("backup", "invalid", "master", "running"):
                if n not in result:
                    continue

                setattr(self, n, result[n])
                yield nagiosplugin.Metric(
                    name=n,
                    value=result[n],
                )


class InterfaceVrrpDisabled(BooleanContext):
    def evaluate(self, metric, resource: InterfaceVrrpCheck):
        if metric.value is True:
            return self.result_cls(nagiosplugin.state.Warn, "VRRP is disabled", metric)
        return self.result_cls(nagiosplugin.state.Ok)


class InterfaceVrrpInvalid(BooleanContext):
    def evaluate(self, metric, resource: InterfaceVrrpCheck):
        if metric.value is True:
            return self.result_cls(
                state=nagiosplugin.state.Warn,
                hint="VRRP config is invalid"
            )
        return self.result_cls(nagiosplugin.state.Ok)


class InterfaceVrrpMaster(BooleanContext):
    def evaluate(self, metric, resource: InterfaceVrrpCheck):
        if not metric.value and resource.master_must:
            return self.result_cls(
                state=nagiosplugin.state.Warn,
                hint="VRRP interface is not master"
            )
        return self.result_cls(nagiosplugin.state.Ok)


@cli.command("interface.vrrp")
@click.option(
    "--name",
    required=True,
    help="The name of the VRRP interface to check",
)
@click.option(
    "--master",
    default=False,
    help="If set the interface must be master",
)
@click.pass_context
def interface_vrrp(ctx, name, master):
    """Check the state of VRRP interfaces"""
    check = nagiosplugin.Check()

    check.add(
        InterfaceVrrpCheck(
            cmd_options=ctx.obj,
            check=check,
            name=name,
            master_must=master,
        ),
        BooleanContext("backup"),
        InterfaceVrrpDisabled("disabled"),
        InterfaceVrrpInvalid("invalid"),
        InterfaceVrrpMaster("master"),
        BooleanContext("running")
    )

    check.main(verbose=ctx.obj["verbose"])
